/*
 * gobject-list: a LD_PRELOAD library for tracking the lifetime of GObjects
 *
 * Copyright (C) 2011 Collabora Ltd.
 * Copyright (C) 2011 Morten Welinder
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
 * USA
 *
 * Authors:
 *     Danielle Madeley  <danielle.madeley@collabora.co.uk>
 *     Morten Welinder <terra@gnome.org>
 */

/* Minimal Makefile:

   CFLAGS=`pkg-config --cflags glib-2.0`
   LIBS=`pkg-config --libs glib-2.0`

   libgobject-list.so: gobject-list.c
   gcc -fPIC -rdynamic -g -c -Wall ${CFLAGS} $<
   gcc -shared -Wl,-soname,$@ -o $@ gobject-list.o -lc -ldl ${LIBS}
*/

/* We need this to get RTLD_NEXT. */
#define _GNU_SOURCE 1

#include <glib-object.h>
#include <gobject/gvaluecollector.h>
#include <gio/gio.h>

#include <dlfcn.h>
#include <signal.h>
#include <string.h>
#include <stdlib.h>
#include <execinfo.h>
#include <stdio.h>
#include <malloc.h>
#include <unistd.h>

static gboolean executing_python_binary = FALSE;
static gboolean display_instance_creations = FALSE;

G_LOCK_DEFINE_STATIC (objects_lock);
G_LOCK_DEFINE_STATIC (object_tree_lock);
static GHashTable *objects = NULL;
static GNode *root_node = NULL;
static GString *qstr;

static GStaticPrivate current_level_key = G_STATIC_PRIVATE_INIT;
static GStaticPrivate current_nodes_key = G_STATIC_PRIVATE_INIT;

static GObject * (*real_g_object_constructor) (GType                  type,
                                               guint                  n_construct_properties,
                                               GObjectConstructParam *construct_params);
static void (*real_g_object_finalize) (GObject *object);

static void* (*real_dlopen) (const char* filename, int mode);

static gpointer (*real_g_object_newv) (GType object_type, guint n_params,
    GParameter *parameters);

static gpointer (*real_gdk_window_get_user_data) (gpointer, gpointer *);
static gpointer (*real_gdk_display_manager_get) (void);
static void (*real_gdk_flush) (void);
static GSList * (*real_gdk_display_manager_list_displays) (gpointer);
static void (*real_gdk_display_close) (gpointer);
static gboolean seen_gdk_window = FALSE;
static GType (*real_gdk_window_object_get_type) (void);

static const char * (*real_gtk_label_get_text) (gpointer);
static const char * (*real_gtk_printer_get_name) (gpointer);
static gpointer (*real_gtk_widget_get_parent) (gpointer);
static const char * (*real_gtk_widget_get_name) (gpointer);
static GSList * (*real_gtk_size_group_get_widgets) (gpointer);
static GType (*real_gtk_action_group_get_type) (void);
static GType (*real_gtk_action_get_type) (void);
static GType (*real_gtk_label_get_type) (void);
static GType (*real_gtk_printer_get_type) (void);
static GType (*real_gtk_printer_option_get_type) (void);
static GType (*real_gtk_table_get_type) (void);
static GType (*real_gtk_widget_get_type) (void);
static GType (*real_gtk_window_get_type) (void);
static GType (*real_gtk_size_group_get_type) (void);

static GType (*real_pango_font_family_get_type) (void);
static const char * (*real_pango_font_family_get_name) (gpointer);

static void initialize (void);

#define MODULE_ARRAY_SIZE 1024
static void *module_array[MODULE_ARRAY_SIZE] = { 0 };
static int module_array_length = 0;

#define BACKTRACE_SIZE 20

struct _ObjectInfo
{
  gpointer data;
  GType type;
  void *backtrace_buffer[BACKTRACE_SIZE];
};

typedef struct _ObjectInfo ObjectInfo;

static const char *
quoted_string (const char *s)
{
  char *es;

  if (!s)
    return "(null)";

  es = g_strescape (s, NULL);
  g_string_truncate (qstr, 0);
  g_string_append_c (qstr, '\"');
  g_string_append (qstr, es);
  g_string_append_c (qstr, '\"');
  g_free (es);

  return qstr->str;
}

static guint object_info_hash (gconstpointer v)
{
  ObjectInfo *i = (ObjectInfo*) v;
  return GPOINTER_TO_UINT (i->data);
}

static gboolean object_info_equal (gconstpointer v1, gconstpointer v2)
{
  ObjectInfo *i1 = (ObjectInfo*) v1;
  ObjectInfo *i2 = (ObjectInfo*) v2;
  return i1->data == i2->data;
}

static gint object_info_find (gconstpointer elem_data, gconstpointer user_data)
{
  ObjectInfo *info = (ObjectInfo*) elem_data;

  if (info->data == user_data) return 0;

  return 1;
}

static ObjectInfo* object_info_new (gpointer object, GType type)
{
  ObjectInfo *info = (ObjectInfo*) g_slice_alloc0 (sizeof (ObjectInfo));
  info->data = object;
  info->type = type;
  return info;
}

static void object_info_free (gpointer info)
{
  g_slice_free (ObjectInfo, info);
}

static const char *
fake_gtk_printer_option_get_name (gpointer obj)
{
  /* We have no properties or getter function, so...  */

  struct _GtkPrinterOption
  {
    GObject parent_instance;

    char *name;
    char *display_text;
    enum { OINK } type;

    char *value;
  
    int num_choices;
    char **choices;
    char **choices_display;
  
    gboolean activates_default;

    gboolean has_conflict;
    char *group;
  };

  return ((struct _GtkPrinterOption *)obj)->name;
}


static gboolean
object_filter (const char *obj_name)
{
  static const char *filter;
  static gboolean inited = FALSE;

  if (!inited) {
    inited = TRUE;
    filter = g_getenv ("GOBJECT_LIST_FILTER");
  }

  if (filter == NULL)
    return TRUE;
  else
    return (strncmp (filter, obj_name, strlen (filter)) == 0);
}

static void
dump_object_list (void)
{
  int i;
  char **strings;
  GList *keys, *l;

  /*
   * We get all the keys at once so we can release the lock right away.
   * Holding the lock across the loop feels risky.
   */
  G_LOCK (objects_lock);
  keys = g_hash_table_get_keys (objects);

  for (l = keys; l; l = l->next)
    {
      ObjectInfo *info = l->data;
      GObject *obj = info->data;
      g_printerr (" - %p, %s: refs=%u",
                  obj, G_OBJECT_TYPE_NAME (obj), obj->ref_count);

      if (real_gtk_action_group_get_type &&
          G_TYPE_CHECK_INSTANCE_TYPE (obj, real_gtk_action_group_get_type ()))
        {
          gchar *name = NULL;

          g_object_get (obj, "name", &name, NULL);
          if (name)
            g_printerr (" name=%s", quoted_string (name));
          g_free (name);
        }

      if (real_gtk_action_get_type &&
          G_TYPE_CHECK_INSTANCE_TYPE (obj, real_gtk_action_get_type ()))
        {
          gchar *name = NULL;

          g_object_get (obj, "name", &name, NULL);
          if (name)
            g_printerr (" name=%s", quoted_string (name));
          g_free (name);
        }

      if (real_gtk_label_get_type &&
          G_TYPE_CHECK_INSTANCE_TYPE (obj, real_gtk_label_get_type ()))
        {
          const char *label = real_gtk_label_get_text (obj);
          g_printerr (" label=%s", quoted_string (label));
        }

      if (real_gtk_printer_get_type &&
          G_TYPE_CHECK_INSTANCE_TYPE (obj, real_gtk_printer_get_type ()))
        {
          const char *name = real_gtk_printer_get_name (obj);
          g_printerr (" name=%s", name);
        }

      if (real_gtk_printer_option_get_type &&
          G_TYPE_CHECK_INSTANCE_TYPE (obj, real_gtk_printer_option_get_type ()))
        {
          const char *name = fake_gtk_printer_option_get_name (obj);
          if (name)
            g_printerr (" name=%s", quoted_string (name));
        }

      if (real_gtk_table_get_type &&
          G_TYPE_CHECK_INSTANCE_TYPE (obj, real_gtk_table_get_type ()))
        {
          int rows, cols;
          g_object_get (obj, "n-rows", &rows, "n-columns", &cols, NULL);
          g_printerr (" size=%dc%dr", cols, rows);
        }

      if (real_gtk_widget_get_type &&
          G_TYPE_CHECK_INSTANCE_TYPE (obj, real_gtk_widget_get_type ()))
        {
          GObject *parent = real_gtk_widget_get_parent (obj);
          const char *name = real_gtk_widget_get_name (obj);

          if (parent)
            g_printerr (" parent=%p", parent);

          if (name && *name)
            g_printerr (" name=%s", quoted_string (name));
        }

      if (real_gtk_window_get_type &&
          G_TYPE_CHECK_INSTANCE_TYPE (obj, real_gtk_window_get_type ()))
        {
          gchar *title = NULL;

          g_object_get (obj, "title", &title, NULL);
          if (title)
            g_printerr (" title=%s", quoted_string (title));
          g_free (title);
        }

      if (real_gtk_size_group_get_type &&
          G_TYPE_CHECK_INSTANCE_TYPE (obj, real_gtk_size_group_get_type ()))
        {
          GSList *l = real_gtk_size_group_get_widgets (obj);
          if (l)
            {
              g_printerr (" widgets=[");
              while (l) {
                g_printerr ("%p", l->data);
                l = l->next;
                if (l)
                  g_printerr (",");
              }
              g_printerr ("]");
            }
        }

      if (real_gdk_window_object_get_type &&
          G_TYPE_CHECK_INSTANCE_TYPE (obj, real_gdk_window_object_get_type ()))
        {
          gpointer user;

          real_gdk_window_get_user_data (obj, &user);
          if (user)
            g_printerr (" user=%p", user);
        }

      if (G_IS_TYPE_MODULE (obj))
        {
          const char *name = G_TYPE_MODULE (obj)->name;
          if (name)
            g_printerr (" name=%s", quoted_string (name));
        }

      if (G_IS_FILE_INFO (obj))
        {
          const char *name = g_file_info_get_name (G_FILE_INFO (obj));
          if (name)
            g_printerr (" name=%s", quoted_string (name));
        }

      if (G_IS_FILE (obj))
        {
          char *uri = g_file_get_uri (G_FILE (obj));
          g_printerr (" uri=%s", quoted_string (uri));
          g_free (uri);
        }

      if (real_pango_font_family_get_type &&
          G_TYPE_CHECK_INSTANCE_TYPE (obj, real_pango_font_family_get_type ()))
        {
          const char *name = real_pango_font_family_get_name (obj);
          if (name)
            g_printerr (" name=%s", quoted_string (name));
        }

      g_printerr ("\n");

      G_LOCK (object_tree_lock);
      GNode *node = g_node_find (root_node, G_IN_ORDER, G_TRAVERSE_ALL, obj);
      guint depth = 0;
      if (node) depth = g_node_depth (node);
      G_UNLOCK (object_tree_lock);

      if (node && depth > 1)
      {
        node = node->parent;
        int i = 0;
        while (node != NULL && node->data != NULL)
        {
          GList *list_node = g_list_find_custom (keys, node->data, object_info_find);
          ObjectInfo *parent_info = list_node ? list_node->data : NULL;

          for (int j=0; j < i; j++) g_printerr ("  ");
          if (parent_info != NULL)
          {
            GObject *parent_obj = parent_info->data;
            g_printerr (" |-- %p %s refs=%d\n", parent_info->data,
                g_type_name (parent_info->type), parent_obj->ref_count);
          }
          else
          {
            g_printerr (" |-- %p [finalized]\n", node->data);
          }

          node = node->parent;
          i++;
        }
      }

      // backtrace where the object was created
      int backtrace_size = 0;
      for (i = 0; i < BACKTRACE_SIZE; i++)
      {
        backtrace_size = i+1;
        if (info->backtrace_buffer[i] == NULL) break;
      }

      strings = backtrace_symbols (info->backtrace_buffer, backtrace_size);
      for (i = 0; i < backtrace_size; i++)
      {
        char *last_slash = g_strrstr (strings[i], "/");
        g_printerr ("    %s\n", last_slash ? last_slash+1 : strings[i]);
      }
      free (strings);

    }
  
  G_UNLOCK (objects_lock);

  g_list_free (keys);
}

static gint compare_num_instances (gconstpointer a, gconstpointer b, gpointer d)
{
  int i, j;
  sscanf ((char*) a, "%d", &i);
  sscanf ((char*) b, "%d", &j);

  return i-j;
}

static void
dump_object_list2 (void)
{
  GList *keys, *l;

  /*
   * We get all the keys at once so we can release the lock right away.
   * Holding the lock across the loop feels risky.
   */
  G_LOCK (objects_lock);
  keys = g_hash_table_get_keys (objects);
  G_UNLOCK (objects_lock);

  GHashTable *objects_by_type = g_hash_table_new (g_str_hash, g_str_equal);

  for (l = keys; l; l = l->next)
    {
      ObjectInfo *info = l->data;
      GObject *obj = info->data;
      gpointer value = g_hash_table_lookup (objects_by_type,
          G_OBJECT_TYPE_NAME (obj));
      gint int_value = GPOINTER_TO_INT (value);
      int_value++;
      g_hash_table_insert (objects_by_type,
          (gpointer) G_OBJECT_TYPE_NAME (obj), GINT_TO_POINTER (int_value));
    }

  GHashTableIter iter;
  gpointer key, value;

  GSequence *seq = g_sequence_new (g_free);

  g_hash_table_iter_init (&iter, objects_by_type);
  while (g_hash_table_iter_next (&iter, &key, &value))
  {
    int num_inst = GPOINTER_TO_INT (value);
    char *s = g_strdup_printf ("%d %s instances", num_inst, (gchar*) key);
    g_sequence_insert_sorted (seq, s, compare_num_instances, NULL);
  }

  GSequenceIter *s_iter = g_sequence_get_begin_iter (seq);
  while (g_sequence_iter_is_end (s_iter) == FALSE)
  {
    g_printerr ("%s\n", (char*) g_sequence_get (s_iter));
    s_iter = g_sequence_iter_next (s_iter);
  }

  g_sequence_free (seq);

  g_list_free (keys);
}

static void
dump_object_counts (void (*dump_func) (const gchar*, guint, gpointer),
    gpointer data)
{
  GList *keys, *l;

  G_LOCK (objects_lock);
  keys = g_hash_table_get_keys (objects);

  GHashTable *objects_by_type = g_hash_table_new (g_str_hash, g_str_equal);

  for (l = keys; l; l = l->next)
    {
      ObjectInfo *info = l->data;
      GObject *obj = info->data;
      gpointer value = g_hash_table_lookup (objects_by_type,
          G_OBJECT_TYPE_NAME (obj));
      gint int_value = GPOINTER_TO_INT (value);
      int_value++;
      g_hash_table_insert (objects_by_type,
          (gpointer) G_OBJECT_TYPE_NAME (obj), GINT_TO_POINTER (int_value));
    }

  G_UNLOCK (objects_lock);

  GHashTableIter iter;
  gpointer key, value;

  g_hash_table_iter_init (&iter, objects_by_type);
  while (g_hash_table_iter_next (&iter, &key, &value))
  {
    int num_inst = GPOINTER_TO_INT (value);
    dump_func ((gchar*) key, num_inst, data);
  }

  g_list_free (keys);
  
}

static void
dump_object_info (void (*dump_func) (ObjectInfo *info, gpointer),
    gpointer data)
{
  GList *keys, *l;

  G_LOCK (objects_lock);
  keys = g_hash_table_get_keys (objects);

  for (l = keys; l; l = l->next)
    {
      ObjectInfo *info = l->data;

      dump_func (info, data);

      /*
      G_LOCK (object_tree_lock);
      GNode *node = g_node_find (root_node, G_IN_ORDER, G_TRAVERSE_ALL, obj);
      guint depth = 0;
      if (node) depth = g_node_depth (node);
      G_UNLOCK (object_tree_lock);

      if (node && depth > 1)
      {
        node = node->parent;
        int i = 0;
        while (node != NULL && node->data != NULL)
        {
          GList *list_node = g_list_find_custom (keys, node->data, object_info_find);
          ObjectInfo *parent_info = list_node ? list_node->data : NULL;

          for (int j=0; j < i; j++) g_printerr ("  ");
          if (parent_info != NULL)
          {
            GObject *parent_obj = parent_info->data;
            g_printerr (" |-- %p %s refs=%d\n", parent_info->data,
                g_type_name (parent_info->type), parent_obj->ref_count);
          }
          else
          {
            g_printerr (" |-- %p [finalized]\n", node->data);
          }

          node = node->parent;
          i++;
        }
      }
      */

    }
  
  G_UNLOCK (objects_lock);

  g_list_free (keys);
}

static void counts_to_variant (const gchar* name, guint num_instances,
    gpointer data)
{
  GVariantBuilder *builder = (GVariantBuilder*) data;

  g_variant_builder_add (builder, "(su)", name, num_instances);
}

static void info_to_variant (ObjectInfo *info, gpointer data)
{
  GVariantBuilder *builder = (GVariantBuilder*) data;

  g_variant_builder_open (builder, g_variant_type_new ("(suax)"));
  g_variant_builder_add (builder, "s", G_OBJECT_TYPE_NAME (info->data));
  g_variant_builder_add (builder, "u", ((GObject*)(info->data))->ref_count);
  // backtrace where the object was created
  g_variant_builder_open (builder, g_variant_type_new ("ax"));
  int i, backtrace_size = 0;
  for (i = 0; i < BACKTRACE_SIZE; i++)
  {
    backtrace_size = i+1;
    if (info->backtrace_buffer[i] == NULL) break;
    g_variant_builder_add (builder, "x", (gint64) info->backtrace_buffer[i]);
  }
  g_variant_builder_close (builder);
  g_variant_builder_close (builder);
/*
  strings = backtrace_symbols (info->backtrace_buffer, backtrace_size);
  for (i = 0; i < backtrace_size; i++)
  {
    char *last_slash = g_strrstr (strings[i], "/");
    g_printerr ("    %s\n", last_slash ? last_slash+1 : strings[i]);
  }
  free (strings);
*/
}

static gboolean
dump_idle_handler (gpointer data)
{
  if (data == NULL)
  {
    // dump object counts
    GVariantBuilder *builder = g_variant_builder_new (
        g_variant_type_new ("(ia(su))"));
    
    g_variant_builder_add (builder, "i", (int) getpid ());
    g_variant_builder_open (builder, g_variant_type_new ("a(su)"));
    dump_object_counts (counts_to_variant, builder);
    g_variant_builder_close (builder);

    GVariant *v = g_variant_builder_end (builder);
    g_variant_builder_unref (builder);

    GDBusConnection *conn = g_bus_get_sync (G_BUS_TYPE_SESSION, NULL, NULL);
    g_dbus_connection_emit_signal (conn, NULL, "/org/debug/leak/detector",
        "org.debug.LeakDetector", "ObjectCountsSnapshot", v, NULL);
    g_dbus_connection_flush (conn, NULL, NULL, NULL);
    g_object_unref (conn);
  }
  else
  {
    // dump object creation backtraces
    GVariantBuilder *builder = g_variant_builder_new (
        g_variant_type_new ("(ia(suax))"));

    g_variant_builder_add (builder, "i", (int) getpid ());
    g_variant_builder_open (builder, g_variant_type_new ("a(suax)"));
    dump_object_info (info_to_variant, builder);
    g_variant_builder_close (builder);
    
    GVariant *v = g_variant_builder_end (builder);
    g_variant_builder_unref (builder);

    GDBusConnection *conn = g_bus_get_sync (G_BUS_TYPE_SESSION, NULL, NULL);
    g_dbus_connection_emit_signal (conn, NULL, "/org/debug/leak/detector",
        "org.debug.LeakDetector", "ObjectsSnapshot", v, NULL);
    g_dbus_connection_flush (conn, NULL, NULL, NULL);
    g_object_unref (conn);
  }

  return FALSE;
}

static void
sig_usr1_handler (int signal)
{
  g_idle_add (dump_idle_handler, NULL);
}

static void
sig_usr2_handler (int signal)
{
  g_idle_add (dump_idle_handler, GINT_TO_POINTER (1));
}

static void
exiting (void)
{
  if (seen_gdk_window && !executing_python_binary)
    {
      GSList *displays;

      real_gdk_flush ();
      while (g_main_context_iteration (NULL, FALSE))
        ;/* nothing */

      displays = real_gdk_display_manager_list_displays
        (real_gdk_display_manager_get ());
      g_slist_foreach (displays, (GFunc)real_gdk_display_close, NULL);
      g_slist_free (displays);
    }

  g_printerr ("\nStill Alive:\n");
  dump_object_list2 ();
/*
  g_printerr ("\n============\n");
  dump_object_list ();
*/
}

static void *
get_func1 (const char *func_name)
{
  return dlsym (RTLD_NEXT, func_name);
}

static void *
get_func2 (const char *func_name)
{
  for (int i = 0; i < module_array_length; i++)
  {
    void *result = dlsym (module_array[i], func_name);
    if (result) return result;
  }

  g_printerr ("Unable to find \"%s\"\n", func_name);
  return NULL;
}

void *
dlopen (const char *file, int mode)
{
  if (!real_dlopen)
  {
    real_dlopen = get_func1 ("dlopen");
  }
  void *result = real_dlopen (file, mode);
  if (result)
  {
    module_array[module_array_length++] = result;
  }

  return result;
}

static GObject*
hook_g_object_constructor (GType                  type,
                           guint                  n_construct_properties,
                           GObjectConstructParam *construct_params)
{
  GObject *obj = real_g_object_constructor (type,
                                            n_construct_properties,
                                            construct_params);

  if (obj)
  {
    // FIXME: proper parent
    int current_level = GPOINTER_TO_INT (g_static_private_get (&current_level_key));
    GNode **parents = ((GNode**) g_static_private_get (&current_nodes_key));
    GNode *parent = parents[current_level-1];
    G_LOCK (object_tree_lock);
    GNode *child = g_node_append_data (parent ? parent : root_node, obj);
    G_UNLOCK (object_tree_lock);
    parents[current_level] = child;
  }

  return obj;
}

static void
hook_g_object_finalize (GObject *obj)
{
  const char *typ = G_OBJECT_TYPE_NAME (obj);
  ObjectInfo *dummy = object_info_new (obj, G_TYPE_INVALID);
  G_LOCK (objects_lock);
  g_hash_table_remove (objects, dummy);
  G_UNLOCK (objects_lock);
  object_info_free (dummy);
  real_g_object_finalize (obj);
  if (display_instance_creations)
    g_printerr (" -- Finalized object %p, %s\n", obj, typ);
}

GType
gdk_window_object_get_type (void)
{
  seen_gdk_window = TRUE;
  if (!real_gdk_window_object_get_type)
  {
    real_gdk_window_object_get_type = get_func2 ("gdk_window_object_get_type");
    // get all the other functions
    real_gdk_window_get_user_data = get_func2 ("gdk_window_get_user_data");
    real_gdk_display_manager_get = get_func2 ("gdk_display_manager_get");
    real_gdk_display_manager_list_displays = get_func2 ("gdk_display_manager_list_displays");
    real_gdk_display_close = get_func2 ("gdk_display_close");
    real_gdk_flush = get_func2 ("gdk_flush");

    real_pango_font_family_get_type = get_func2 ("pango_font_family_get_type");
    real_pango_font_family_get_name = get_func2 ("pango_font_family_get_name");
  }
  return real_gdk_window_object_get_type ();
}

void
g_type_init (void)
{
  initialize ();
}

GObject*
g_object_new_valist (GType        object_type,
                     const gchar *first_property_name,
                     va_list      var_args)
{
  GObjectClass *class;
  GParameter *params;
  const gchar *name;
  GObject *object;
  guint n_params = 0, n_alloced_params = 16;

  g_return_val_if_fail (G_TYPE_IS_OBJECT (object_type), NULL);

  if (!first_property_name)
    return g_object_newv (object_type, 0, NULL);

  class = g_type_class_ref (object_type);

  params = g_new0 (GParameter, n_alloced_params);
  name = first_property_name;
  while (name)
    {
      gchar *error = NULL;
      GParamSpec *pspec = g_object_class_find_property (class, name);

      if (!pspec)
        {
          g_warning ("%s: object class `%s' has no property named `%s'",
                     G_STRFUNC,
                     g_type_name (object_type),
                     name);
          break;
        }
      if (n_params >= n_alloced_params)
        {
          n_alloced_params += 16;
          params = g_renew (GParameter, params, n_alloced_params);
          memset (params + n_params, 0, 16 * (sizeof *params));
        }
      params[n_params].name = name;
      G_VALUE_COLLECT_INIT (&params[n_params].value, pspec->value_type,
                            var_args, 0, &error);
      if (error)
        {
          g_warning ("%s: %s", G_STRFUNC, error);
          g_free (error);
          g_value_unset (&params[n_params].value);
          break;
        }
      n_params++;
      name = va_arg (var_args, gchar*);
    }

  object = g_object_newv (object_type, n_params, params);

  while (n_params--)
    g_value_unset (&params[n_params].value);
  g_free (params);

  g_type_class_unref (class);

  return object;
}

gpointer
g_object_new (GType object_type, const gchar *first_prop, ...)
{
  gpointer object;
  va_list var_args;
  
  if (!first_prop) return g_object_newv (object_type, 0, NULL);

  va_start (var_args, first_prop);
  object = g_object_new_valist (object_type, first_prop, var_args);
  va_end (var_args);

  return object;
}

gpointer
g_object_newv (GType object_type, guint n_params, GParameter *parameters)
{
  const char *obj_name;

  if (g_static_private_get (&current_nodes_key) == NULL)
  {
    g_static_private_set (&current_nodes_key, g_new0 (void*, 128), g_free);
  }

  gint current_level = GPOINTER_TO_INT (g_static_private_get (&current_level_key));
  current_level++;
  g_static_private_set (&current_level_key, GINT_TO_POINTER (current_level), NULL);

  /* ==== REAL CONSTRUCTION ==== */
  gpointer object = real_g_object_newv (object_type, n_params, parameters);
  /* ==== CONSTRUCTION END ==== */

  current_level--;
  g_static_private_set (&current_level_key, GINT_TO_POINTER (current_level), NULL);

  GNode **parents = (GNode**) g_static_private_get (&current_nodes_key);
  parents[current_level+1] = NULL;

  if (!object) return NULL;

  obj_name = G_OBJECT_TYPE_NAME (object);
  if (object_filter (obj_name))
    {
      if (display_instance_creations)
      {
        g_printerr (" ++ Created object %p, %s [%d]\n", object, obj_name, current_level);
      }
      ObjectInfo *info = object_info_new (object, G_OBJECT_TYPE (object));
      backtrace (info->backtrace_buffer, BACKTRACE_SIZE);

      G_LOCK (objects_lock);
      g_hash_table_insert (objects, info, GUINT_TO_POINTER (TRUE));
      G_UNLOCK (objects_lock);
    }

  return object;
}


static void
initialize (void)
{
  if (G_LIKELY (objects != NULL))
    return;

  /* Note: we are not in a sane state until we call the real g_type_init().  */

  {
    void (*real_g_type_init) (void) = get_func1 ("g_type_init");
    real_g_type_init ();

    // let's check if this is python
    void *py_init = dlsym (RTLD_DEFAULT, "Py_Initialize");
    executing_python_binary = py_init != NULL;
  }

  /* set up signal handler */
  signal (SIGUSR1, sig_usr1_handler);
  signal (SIGUSR2, sig_usr2_handler);

  /* set up objects map */
  objects = g_hash_table_new_full (object_info_hash, object_info_equal,
                                   object_info_free, NULL);
  root_node = g_node_new (NULL);

  qstr = g_string_new (NULL);

  /* Prime things while we're still single-threaded.  */
  object_filter ("oink");

  /* Set up exit handler */
  atexit (exiting);

  {
    GObjectClass *klass = g_type_class_ref (G_TYPE_OBJECT);

    real_g_object_constructor = klass->constructor;
    klass->constructor = hook_g_object_constructor;

    real_g_object_finalize = klass->finalize;
    klass->finalize = hook_g_object_finalize;

    g_type_class_unref (klass);
  }

  real_g_object_newv = get_func1 ("g_object_newv");
  real_gdk_window_object_get_type = get_func1 ("gdk_window_object_get_type");

  real_gdk_window_get_user_data = get_func1 ("gdk_window_get_user_data");
  real_gdk_display_manager_get = get_func1 ("gdk_display_manager_get");
  real_gdk_display_manager_list_displays = get_func1 ("gdk_display_manager_list_displays");
  real_gdk_display_close = get_func1 ("gdk_display_close");
  real_gdk_flush = get_func1 ("gdk_flush");

  real_gtk_action_get_type = get_func1 ("gtk_action_get_type");
  real_gtk_action_group_get_type = get_func1 ("gtk_action_group_get_type");
  real_gtk_label_get_type = get_func1 ("gtk_label_get_type");
  real_gtk_printer_get_type = get_func1 ("gtk_printer_get_type");
  real_gtk_printer_option_get_type = get_func1 ("gtk_printer_option_get_type");
  real_gtk_table_get_type = get_func1 ("gtk_table_get_type");
  real_gtk_widget_get_type = get_func1 ("gtk_widget_get_type");
  real_gtk_window_get_type = get_func1 ("gtk_window_get_type");
  real_gtk_size_group_get_type = get_func1 ("gtk_size_group_get_type");
  real_gtk_label_get_text = get_func1 ("gtk_label_get_text");
  real_gtk_printer_get_name = get_func1 ("gtk_printer_get_name");
  real_gtk_widget_get_parent = get_func1 ("gtk_widget_get_parent");
  real_gtk_widget_get_name = get_func1 ("gtk_widget_get_name");
  real_gtk_size_group_get_widgets = get_func1 ("gtk_size_group_get_widgets");

  real_pango_font_family_get_type = get_func1 ("pango_font_family_get_type");
  real_pango_font_family_get_name = get_func1 ("pango_font_family_get_name");

  display_instance_creations = g_getenv ("DEBUG_INSTANCES") != NULL;
}
